@@ -1,11 +1,12 @@
-;;; cflow.el --- major mode for viewing cflow output files.
+;;; cflow.el --- major mode for viewing cflow output files.  -*- lexical-binding: t; -*-
 
 ;; Authors: 1994, 1995, 2005, 2007, 2021 Sergey Poznyakoff
 ;; Version: 0.2.1
 ;; Keywords: cflow
 
 ;; This file is part of GNU cflow
-;; Copyright (C) 1994-2024 Sergey Poznyakoff
+;; Copyright (C) 1994-2025 Sergey Poznyakoff
+;; Copyright (C) 2025  Free Software Foundation, Inc.
  
 ;; GNU cflow is free software; you can redistribute it and/or modify
 ;; it under the terms of the GNU General Public License as published by
@@ -20,7 +21,10 @@
 ;; You should have received a copy of the GNU General Public License
 ;; along with this program.  If not, see <http://www.gnu.org/licenses/>.
 
-;; Installation:
+;;; Commentary:
+
+;;;; Installation:
+
 ;; You may wish to use precompiled version of the module. To create it
 ;; run:
 ;;    emacs -batch -f batch-byte-compile cflow-mode.el
@@ -30,47 +34,49 @@
 ;; Customization:
 ;;  To your .emacs or site-start.el add:
 ;;  (autoload 'cflow-mode "cflow-mode")
-;;  (setq auto-mode-alist (append auto-mode-alist
-;;                                '(("\\.cflow$" . cflow-mode))))
+;;  (add-to-list 'auto-mode-alist '("\\.cflow\\'" . cflow-mode))
 
-(eval-when-compile
-  ;; We use functions from these modules
-  (mapcar 'require '(font-lock)))
+;;; Code:
 
-(defvar cflow-mode-syntax-table nil
-  "Syntax table used in cflow-mode buffers.")
+(require 'font-lock) ;; FIXME: Not needed any more, I think.
 
-(unless cflow-mode-syntax-table
-  (setq cflow-mode-syntax-table (make-syntax-table))
-  (modify-syntax-entry ?\# "<" cflow-mode-syntax-table)
-  (modify-syntax-entry ?\n ">" cflow-mode-syntax-table))
+(defvar cflow-mode-syntax-table
+  (let ((st (make-syntax-table)))
+    (modify-syntax-entry ?\# "<" st)
+    (modify-syntax-entry ?\n ">" st)
+    st))
 
   
-(defvar cflow-mode-map (make-sparse-keymap)
+(defvar cflow-mode-map
+  (let ((map (make-sparse-keymap)))
+    (define-key map "s" #'cflow-find-function)
+    (define-key map "r" #'cflow-recursion-root)
+    (define-key map "R" #'cflow-recursion-next)
+    (define-key map "x" #'cflow-goto-expand)
+    (define-key map "E" #'cflow-edit-out-full)
+    (define-key map "c" #'cflow-find-caller)
+    (define-key map "n" #'cflow-next-function)
+    (define-key map "p" #'cflow-prev-function)
+
+    (define-key map [menu-bar] (make-sparse-keymap))
+
+    (define-key map [menu-bar cflow]
+                (cons "Cflow" (make-sparse-keymap "Cflow")))
+
+    (define-key map [menu-bar cflow cflow-recursion-next]
+                '("Recursion next" . cflow-recursion-next))
+    (define-key map [menu-bar cflow cflow-recursion-root]
+                '("Recursion root" . cflow-recursion-root))
+    (define-key map [menu-bar cflow cflow-goto-expand]
+                '("Find expansion" . cflow-goto-expand))
+    (define-key map [menu-bar cflow cflow-find-function]
+                '("Find function" . cflow-find-function))
+    map)
   "Keymap used in Cflow mode.")
 
-(define-key cflow-mode-map "s" 'cflow-find-function)
-(define-key cflow-mode-map "r" 'cflow-recursion-root)
-(define-key cflow-mode-map "R" 'cflow-recursion-next)
-(define-key cflow-mode-map "x" 'cflow-goto-expand)
-(define-key cflow-mode-map "E" 'cflow-edit-out-full)
-(define-key cflow-mode-map "c" 'cflow-find-caller)
-(define-key cflow-mode-map "n" 'cflow-next-function)
-(define-key cflow-mode-map "p" 'cflow-prev-function)
-
-(define-key cflow-mode-map [menu-bar] (make-sparse-keymap))
-
-(define-key cflow-mode-map [menu-bar cflow]
-  (cons "Cflow" (make-sparse-keymap "Cflow")))
-
-(define-key cflow-mode-map [menu-bar cflow cflow-recursion-next]
-  '("Recursion next" . cflow-recursion-next))
-(define-key cflow-mode-map [menu-bar cflow cflow-recursion-root]
-  '("Recursion root" . cflow-recursion-root))
-(define-key cflow-mode-map [menu-bar cflow cflow-goto-expand]
-  '("Find expansion" . cflow-goto-expand))
-(define-key cflow-mode-map [menu-bar cflow cflow-find-function]
-  '("Find function" . cflow-find-function))
+(defun cflow--goto-line (n)
+  (goto-char (point-min))
+  (forward-line (1- n)))
 
 ;; Find the function under cursor.
 ;; Switch to the proper buffer and go to the function header
@@ -80,7 +86,7 @@
     (cond
      (lst
       (switch-to-buffer (find-file-noselect (car lst)))
-      (goto-line (car (cdr lst))))
+      (cflow--goto-line (car (cdr lst))))
      (t
       (error "No source/line information for this function.")))))
 
@@ -117,10 +123,13 @@
     (cond
      ((> num 0)
       (push-mark)
-      (goto-line num))
+      (cflow--goto-line num))
      (t
       (error "Not a recursive call")))))
 
+
+(defvar-local cflow-recursion-root-line nil)
+
 (defun cflow-recursion-next ()
   "Go to next recursive call"
   (interactive)
@@ -129,13 +138,13 @@
     (cond
      ((re-search-forward "(R)"
 			 (save-excursion (end-of-line) (point)) t)
-      (setq cflow-recursion-root-line (count-lines 1 (point))))))
+      (setq cflow-recursion-root-line (count-lines (point-min) (point))))))
   (cond
    ((null cflow-recursion-root-line)
     (error "No recursive functions"))
    (t
     (let ((pos (save-excursion
-		 (next-line 1)
+		 (forward-line 1)
 		 (re-search-forward
 		  (concat "(recursive: see "
 			  (number-to-string cflow-recursion-root-line)
@@ -144,7 +153,7 @@
 		  t))))
       (cond
        ((null pos)
-	(goto-line cflow-recursion-root-line)
+	(cflow--goto-line cflow-recursion-root-line)
 	(error "No more calls."))
        (t
 	(push-mark)
@@ -166,7 +175,7 @@
     (cond
      ((> num 0)
       (push-mark)
-      (goto-line num))
+      (cflow--goto-line num))
      (t
       (error "Not expandable")))))
 
@@ -201,8 +210,7 @@
 	       (beginning-of-line)
 	       (cond
 		((looking-at "^\\( +\\)\\S-")
-		 (let ((indent (- (match-end 1) (match-beginning 1)))
-		       (rx (format "^%s\\S-" (match-string 0))))
+		 (let ((indent (- (match-end 1) (match-beginning 1))))
 		   (let ((pos nil)
 			 (stop (if (> advance 0) 'eobp 'bobp)))
 		     (while (not pos)
@@ -236,53 +244,40 @@
   (interactive "P")
   (cflow-goto-sibling (or arg 1) -1))
 
-(defvar cflow-read-only)
+(defvar cflow--read-only)
 
 (defun cflow-edit-out-full ()
   "Get out of Cflow mode, leaving Cflow file buffer in fundamental mode."
   (interactive)
   (if (yes-or-no-p "Should I let you edit the whole Cflow file? ")
       (progn
-	(setq buffer-read-only cflow-read-only)
+	(setq buffer-read-only cflow--read-only)
 	(fundamental-mode)
 	(message "Type 'M-x cflow-mode RET' once done"))))
 
 
 ;; Font-lock stuff
 (defconst cflow-font-lock-keywords
-  (eval-when-compile
-    (list
-     (cons "^\\s *[0-9]+" font-lock-constant-face)
-     (list "\\(\\S +\\)()\\s +\\(<[^>]*>\\)"
-	   '(1 font-lock-function-name-face)
-	   '(2 font-lock-type-face))
-     (list "\\(\\S +\\)\\s +\\(<[^>]*>\\)"
-	   '(1 font-lock-variable-name-face)
-	   '(2 font-lock-type-face))
-     (cons "\\S +()$" font-lock-builtin-face)
-     (cons "(R):?$" font-lock-comment-face)
-     (cons "(recursive: see [0-9]+)" font-lock-comment-face)
-     (cons "^[ \\t+-|\\]+" font-lock-keyword-face))))
+  `(("^ *[0-9]+" (0 'font-lock-constant-face))
+    ("\\([^ \t\n]+\\)() +\\(<[^>]*>\\)"
+     (1 'font-lock-function-name-face)
+     (2 'font-lock-type-face))
+    ("\\([^ \t\n]+\\) +\\(<[^>]*>\\)"
+     (1 'font-lock-variable-name-face)
+     (2 'font-lock-type-face))
+    ("[^ \t\n]+()$" (0 'font-lock-builtin-face))
+    ("\\((R):?\\)\\(?: \\[see [0-9]+]\\)?$" (1 'font-lock-comment-face))
+    ("(recursive: see [0-9]+)" (0 'font-lock-comment-face))
+    ("\\[see [0-9]+]" (0 'font-lock-comment-face))
+    ("^[- \\t+|\\]+" (0 'font-lock-keyword-face))))
 
 ;;;###autoload
-(defun cflow-mode ()
-  "Major mode for viewing cflow output files
+(define-derived-mode cflow-mode nil "Cflow"
+  "Major mode for viewing cflow output files."
 
-Key bindings are:
-\\{cflow-mode-map}
-"
-  (interactive)
-  (kill-all-local-variables)
-  (use-local-map cflow-mode-map)
-  (setq major-mode 'cflow-mode
-	mode-name "Cflow")
-  (make-variable-buffer-local 'cflow-recursion-root-line)
-
-  (set (make-local-variable 'cflow-read-only) buffer-read-only)
+  (set (make-local-variable 'cflow--read-only) buffer-read-only)
   (setq buffer-read-only t)
 
-  (set-default 'cflow-recursion-root-line nil)
-
   (set-syntax-table cflow-mode-syntax-table)
 
   (make-local-variable 'font-lock-defaults)
@@ -291,6 +286,9 @@ Key bindings are:
 	  (("+-*/.<>=!?$%_&~^:" . "w"))
 	  beginning-of-line)))
 
+;;;###autoload
+(add-to-list 'auto-mode-alist '("\\.cflow\\'" . cflow-mode))
+
 (provide 'cflow-mode)
 ;;; cflow-mode ends
 
