@@ -1,11 +1,11 @@
 /* This file is part of GNU cflow
-   Copyright (C) 1997-2021 Sergey Poznyakoff
- 
+   Copyright (C) 1997-2025 Sergey Poznyakoff
+
    GNU cflow is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 3 of the License, or
    (at your option) any later version.
- 
+
    GNU cflow is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
@@ -34,18 +34,14 @@
 #define _(c) gettext(c)
 #define N_(c) c
 
-#if HAVE_LOCALE_H
-# include <locale.h>
-#endif
-#if !HAVE_SETLOCALE
-# define setlocale(category, locale) /* empty */
-#endif
+#include <locale.h>
 
 /* Exit codes */
 #define EX_OK    0  /* Success */
 #define EX_FATAL 1  /* Fatal error */
 #define EX_SOFT  2  /* Some input files cannot be read or parsed */
 #define EX_USAGE 3  /* Command line usage error */
+#define EX_SOFTWARE 4
 
 #define NUMITEMS(a) sizeof(a)/sizeof((a)[0])
 
@@ -99,7 +95,7 @@ struct symbol {
      struct table_entry *owner;
      Symbol *next;                 /* Next symbol with the same hash */
      struct linked_list_entry *entry;
-     
+
      enum symtype type;            /* Type of the symbol */
      char *name;                   /* Identifier */
      enum symbol_flag flag;        /* Specific flag */
@@ -107,7 +103,7 @@ struct symbol {
 				      type==SymToken and flag==symbol_alias.
 				      In this case, the rest of the structure
 				      is ignored */
-	
+
      int active;                   /* Set to 1 when the symbol's subtree is
 				      being processed, prevent recursion */
      int expand_line;              /* Number of the output line where this
@@ -116,21 +112,21 @@ struct symbol {
      int visible;                  /* Visibility marker.  Symbol will be
 				      shown on output only if its value equals
 				      that of the output_visible global. */
-     
+
      int token_type;               /* Type of the token */
      char *source;                 /* Source file */
      int def_line;                 /* Source line */
      struct linked_list *ref_line; /* Referenced in */
-     
+
      int level;                    /* Block nesting level (for local vars),
 				      Parameter nesting level (for params) */
-     
-     char *decl;                   /* Declaration */ 
+
+     char *decl;                   /* Declaration */
      enum storage storage;         /* Storage type */
-     
+
      int arity;                    /* Number of parameters or -1 for
-				      variables */  
-     
+				      variables */
+
      int recursive;                /* Is the function recursive */
      size_t ord;                   /* ordinal number */
      struct linked_list *caller;   /* List of callers */
@@ -168,15 +164,15 @@ extern int reverse_tree;
 extern int out_line;
 extern char *start_name;
 extern int all_functions;
-extern int max_depth;
+extern unsigned max_depth;
 extern int emacs_option;
 extern int debug;
 extern int preprocess_option;
 extern int omit_arguments_option;
 extern int omit_symbol_names_option;
 
-extern int token_stack_length;
-extern int token_stack_increase;
+extern unsigned token_stack_length;
+extern unsigned token_stack_increase;
 
 extern int symbol_count;
 extern unsigned input_file_count;
@@ -207,13 +203,13 @@ void delete_autos(int level);
 void delete_statics(void);
 void delete_parms(int level);
 void move_parms(int level);
-size_t collect_symbols(Symbol ***, int (*sel)(), size_t rescnt);
+size_t collect_symbols(Symbol ***, int (*sel)(Symbol *p), size_t rescnt);
 size_t collect_functions(Symbol ***return_sym);
 struct linked_list *linked_list_create(linked_list_free_data_fp fun);
 void linked_list_destroy(struct linked_list **plist);
 void linked_list_append(struct linked_list **plist, void *data);
 void linked_list_prepend(struct linked_list **plist, void *data);
-void linked_list_iterate(struct linked_list **plist, 
+void linked_list_iterate(struct linked_list **plist,
 			 int (*itr) (void *, void *), void *data);
 void linked_list_unlink(struct linked_list *list,
 			struct linked_list_entry *ent);
@@ -225,7 +221,7 @@ int get_token(void);
 int source(char *name);
 void init_lex(int debug_level);
 void set_preprocessor(const char *arg);
-void pp_option(const char *arg); 
+void pp_option(int opt, const char *arg);
 
 void init_parse(void);
 int yyparse(void);
