@@ -1,11 +1,11 @@
 /* This file is part of GNU cflow
-   Copyright (C) 1997-2019 Sergey Poznyakoff
- 
+   Copyright (C) 1997-2025 Sergey Poznyakoff
+
    GNU cflow is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 3 of the License, or
    (at your option) any later version.
- 
+
    GNU cflow is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
@@ -26,6 +26,7 @@ typedef struct {
      enum storage storage;
 } Ident;
 
+void skip_declaration(void);
 void parse_declaration(Ident*, int);
 void parse_variable_declaration(Ident*, int);
 void parse_function_declaration(Ident*, int);
@@ -43,15 +44,15 @@ int dirdcl(Ident*);
 void skip_struct();
 Symbol *get_symbol(char *name);
 void maybe_parm_list(int *parm_cnt_return);
-    
+
 void call(char*, int);
 void reference(char*, int);
 
-int level;                  /* Current nesting level */
-Symbol *caller;             /* Current caller */
-struct obstack text_stk;    /* Obstack for composing declaration line */
+static int level;                  /* Current nesting level */
+static Symbol *caller;             /* Current caller */
+static struct obstack text_stk;    /* Obstack for composing declaration line */
 
-int parm_level;             /* Parameter declaration nesting level */
+static int parm_level;             /* Parameter declaration nesting level */
 
 typedef struct {
      int type;
@@ -61,12 +62,12 @@ typedef struct {
 
 typedef int Stackpos[1];
 
-TOKSTK tok;
-TOKSTK *token_stack;
-int tos;
-int curs;
-int token_stack_length = 64;
-int token_stack_increase = 32;
+static TOKSTK tok;
+static TOKSTK *token_stack;
+static int tos;
+static int curs;
+unsigned token_stack_length = 64;
+unsigned token_stack_increase = 32;
 static int need_space;
 
 void mark(Stackpos);
@@ -158,7 +159,7 @@ token_type_str(int t)
 	  return "PARM_WRAPPER";
      case QUALIFIER:
 	  return "QUALIFIER";
-     }	  
+     }
      if (isprint(t))
 	  snprintf(buf, sizeof(buf), "'%c'(%d)", t, t);
      else
@@ -183,7 +184,7 @@ debugtoken(TOKSTK *t, char *fmt, ...)
      if (debug > 1) {
 	  va_list ap;
 	  int i;
-	  
+
 	  if (fmt) {
 	       va_start(ap, fmt);
 	       vfprintf(stderr, fmt, ap);
@@ -275,7 +276,7 @@ cleanup_stack()
 {
      int delta = tos - curs;
 
-     if (delta > 0) 
+     if (delta > 0)
 	  memmove(token_stack, token_stack+curs, delta*sizeof(token_stack[0]));
      else if (delta < 0) /* Invalid input */
 	  delta = 0;
@@ -293,7 +294,7 @@ int
 nexttoken()
 {
      int type;
-     
+
      if (curs == tos) {
 	  type = get_token();
 	  tokpush(type, line_num, yylval.str);
@@ -333,7 +334,7 @@ void
 save_token(TOKSTK *tokptr)
 {
      int len;
-    
+
      switch (tokptr->type) {
      case IDENTIFIER:
      case TYPE:
@@ -341,16 +342,16 @@ save_token(TOKSTK *tokptr)
      case PARM_WRAPPER:
      case WORD:
      case QUALIFIER:
-	  if (need_space) 
+	  if (need_space)
 	       obstack_1grow(&text_stk, ' ');
 	  len = strlen(tokptr->token);
 	  obstack_grow(&text_stk, tokptr->token, len);
 	  need_space = 1;
 	  break;
      case MODIFIER:
-	  if (need_space) 
+	  if (need_space)
 	       obstack_1grow(&text_stk, ' ');
-	  if (tokptr->token[0] == '*') 
+	  if (tokptr->token[0] == '*')
 	       need_space = 0;
 	  else
 	       need_space = 1;
@@ -365,7 +366,7 @@ save_token(TOKSTK *tokptr)
 	  need_space = 1;
 	  break;
      case '(':
-	  if (need_space) 
+	  if (need_space)
 	       obstack_1grow(&text_stk, ' ');
 	  obstack_1grow(&text_stk, tokptr->type);
 	  need_space = 0;
@@ -381,14 +382,14 @@ save_token(TOKSTK *tokptr)
 	  break;
      case LBRACE:
      case LBRACE0:
-	  if (need_space) 
+	  if (need_space)
 	       obstack_1grow(&text_stk, ' ');
 	  obstack_1grow(&text_stk, '{');
 	  need_space = 1;
 	  break;
      case RBRACE:
      case RBRACE0:
-	  if (need_space) 
+	  if (need_space)
 	       obstack_1grow(&text_stk, ' ');
 	  obstack_1grow(&text_stk, '}');
 	  need_space = 1;
@@ -425,7 +426,7 @@ save_stack_is_empty()
 {
      return save_end <= 0;
 }
-	  
+
 char *
 finish_save_stack(char *name)
 {
@@ -445,7 +446,7 @@ finish_save_stack(char *name)
 	       }
 	       break;
 	  case ')':
-	       if (omit_arguments_option) 
+	       if (omit_arguments_option)
 		    level--;
 	       break;
 	  case IDENTIFIER:
@@ -471,9 +472,56 @@ skip_to(int c)
      }
 }
 
-int
-skip_balanced(int open_tok, int close_tok, int level)
+struct balance_state {
+     int idx;
+     int level;
+     struct balance_state *prev;
+};
+
+static void
+push_balance_state(struct balance_state **ptos, int idx, int level)
 {
+     struct balance_state *state = xmalloc(sizeof(state[0]));
+     state->idx = idx;
+     state->level = level;
+     state->prev = *ptos;
+     *ptos = state;
+}
+
+static void
+pop_balance_state(struct balance_state **ptos, int *idx, int *level)
+{
+     struct balance_state *state = *ptos;
+     *idx = state->idx;
+     *level = state->level;
+     *ptos = state->prev;
+     free(state);
+}
+
+static void
+free_balance_stack(struct balance_state **ptos)
+{
+     int i, j;
+     while (*ptos)
+	  pop_balance_state(ptos, &i, &j);
+}
+
+static int
+find_closing_paren(int open_tok, int level)
+{
+     static char oparen[] = "([{";
+     static char cparen[] = ")]}";
+     int idx;
+     char *p = strchr(oparen,  open_tok);
+     struct balance_state *tos = NULL;
+
+     if (!p) {
+	  error(EX_SOFTWARE, 0, "%s:%d: INTERNAL ERROR: malformed call to %s",
+		__FILE__, __LINE__, __func__);
+	  abort();
+     }
+     idx = p - oparen;
+
      if (level == 0) {
 	  if (nexttoken() != open_tok) {
 	       return 1;
@@ -481,23 +529,39 @@ skip_balanced(int open_tok, int close_tok, int level)
 	  level++;
      }
      while (nexttoken()) {
-	  if (tok.type == LBRACE0 && open_tok == '{')
+	  if (tok.type == LBRACE0)
 	       tok.type = '{';
-	  else if (tok.type == RBRACE0 && close_tok == '}')
+	  else if (tok.type == RBRACE0)
 	       tok.type = '}';
-	  
-	  if (tok.type == open_tok) 
+
+	  if (tok.type == oparen[idx])
 	       level++;
-	  else if (tok.type == close_tok) {
+	  else if (tok.type == cparen[idx]) {
 	       if (--level == 0) {
-		    nexttoken();
-		    return 0;
+		    if (tos == NULL)
+			 return 0;
+		    else
+			 pop_balance_state(&tos, &idx, &level);
 	       }
-	  } 
+	  } else if ((p = strchr(oparen, tok.type)) != NULL) {
+	       push_balance_state(&tos, idx, level);
+	       idx = p - oparen;
+	       level = 1;
+	  }
      }
+     free_balance_stack(&tos);
      return -1;
 }
 
+static inline int
+skip_balanced(int open_tok, int level)
+{
+     int rc = find_closing_paren(open_tok, level);
+     if (rc == 0)
+	  nexttoken();
+     return rc;
+}
+
 int
 yyparse()
 {
@@ -516,8 +580,12 @@ yyparse()
 	  case TYPEDEF:
 	       parse_typedef();
 	       break;
+	  case DECLARATION:
+	       skip_declaration();
+	       break;
 	  case EXTERN:
 	       identifier.storage = ExplicitExternStorage;
+	       nexttoken();
 	       parse_declaration(&identifier, 0);
 	       break;
 	  case STATIC:
@@ -554,7 +622,7 @@ is_function()
 	       nexttoken();
 	       continue;
 	  case PARM_WRAPPER:
-	       if (skip_balanced('(', ')', 0) == -1)
+	       if (skip_balanced('(', 0) == -1)
 		    file_error(_("unexpected end of file in declaration"),
 			       NULL);
 	       continue;
@@ -566,7 +634,7 @@ is_function()
 	  }
 	  break;
      }
-     
+
      restore(sp);
      return res;
 }
@@ -574,13 +642,21 @@ is_function()
 void
 parse_declaration(Ident *ident, int parm)
 {
-     if (is_function()) 
+     if (is_function())
 	  parse_function_declaration(ident, parm);
      else
 	  parse_variable_declaration(ident, parm);
      delete_parms(parm_level);
 }
 
+void
+skip_declaration(void)
+{
+     if (skip_balanced('(', 0) == -1)
+	  file_error(_("unexpected end of file in declaration"), NULL);
+     if (tok.type != ';')
+	  putback();
+}
 
 void
 expression()
@@ -594,6 +670,14 @@ expression()
 	  switch (tok.type) {
 	  case ';':
 	       return;
+	  case PARM_WRAPPER:
+	       if (skip_balanced('(', 0) == -1) {
+		    file_error(_("unexpected end of file in expression"),
+			       NULL);
+		    return;
+	       }
+	       putback();
+	       break;
 	  case LBRACE:
 	  case LBRACE0:
 	  case RBRACE:
@@ -609,7 +693,7 @@ expression()
 		    file_error(_("unexpected end of file in expression"),
 			       NULL);
 	       return;
-	    
+
 	  case IDENTIFIER:
 	       name = tok.token;
 	       line = tok.line;
@@ -637,12 +721,23 @@ expression()
 	       break;
 	  case '(':
 	       /* maybe typecast */
-	       if (nexttoken() == TYPE || tok.type == STRUCT)
-		    skip_to(')');
-	       else {
-		    putback();
+	       if (nexttoken() == TYPE || tok.type == STRUCT) {
+		    if (skip_balanced('(', 1) == -1) {
+			 file_error(_("unexpected end of file in expression"),
+				    NULL);
+			 return;
+		    }
+		    if (tok.type == LBRACE || tok.type == LBRACE0) {
+			 if (skip_balanced('{', 1) == -1) {
+			      file_error(_("unexpected end of file in expression"),
+					 NULL);
+			      return;
+			 }
+		    }
+	       } else {
 		    parens_lev++;
 	       }
+	       putback();
 	       break;
 	  case ')':
 	       parens_lev--;
@@ -669,15 +764,15 @@ parse_function_declaration(Ident *ident, int parm)
 	       break;
 	  /*FALLTHROUGH*/
      default:
-	  if (error_recovery) 
+	  if (error_recovery)
 	       nexttoken();
 	  else {
-	       if (verbose) 
+	       if (verbose)
 		    file_error(_("expected `;'"), &tok);
 	       error_recovery = 1;
 	  }
 	  goto restart;
-	  
+
      case ';':
      case ',':
 	  break;
@@ -685,7 +780,8 @@ parse_function_declaration(Ident *ident, int parm)
      case LBRACE:
 	  if (ident->name) {
 	       caller = lookup(ident->name);
-	       if (caller && caller->storage == AutoStorage)
+	       if (caller &&
+		   (caller->storage == AutoStorage || caller->flag == symbol_target))
 		    caller = NULL;
 	       func_body();
 	  }
@@ -699,6 +795,10 @@ parse_function_declaration(Ident *ident, int parm)
 int
 fake_struct(Ident *ident)
 {
+     while (tok.type == QUALIFIER &&
+	    (strcmp(tok.token, "const") == 0 ||
+	     strcmp(tok.token, "volatile") == 0))
+	  nexttoken();
      ident->type_end = -1;
      if (tok.type == STRUCT) {
 	  if (nexttoken() == IDENTIFIER) {
@@ -706,7 +806,16 @@ fake_struct(Ident *ident)
 	  }
 	  putback();
 	  skip_struct();
-	  if (tok.type == IDENTIFIER || tok.type == MODIFIER || tok.type == QUALIFIER) {
+	  if (tok.type == IDENTIFIER || tok.type == MODIFIER ||
+	      tok.type == QUALIFIER)
+	       putback();
+	  else if (tok.type == TYPE) {
+	       /*
+		* Duplicate typedef is allowed as long as both definitions
+		* coincide.  Put the token back as an identifier, so that
+		* dcl parses it correctly.
+		*/
+	       token_stack[curs-1].type = IDENTIFIER;
 	       putback();
 	  } else if (tok.type == '(')
 	       return 0;
@@ -721,9 +830,13 @@ void
 parse_variable_declaration(Ident *ident, int parm)
 {
      Stackpos sp;
-     
+
      mark(sp);
      ident->type_end = -1;
+
+     while (tok.type == QUALIFIER)
+	  nexttoken();
+
      if (tok.type == STRUCT || tok.type == UNION) {
 	  if (nexttoken() == IDENTIFIER) {
 	       ident->type_end = curs;
@@ -749,18 +862,18 @@ parse_variable_declaration(Ident *ident, int parm)
      }
  again:
      parse_dcl(ident, 0);
-     
- select:    
+
+ select:
      switch (tok.type) {
      case ')':
 	  if (parm)
 	       break;
 	  /*FALLTHROUGH*/
      default:
-	  if (verbose) 
+	  if (verbose)
 	       file_error(_("expected `;'"), &tok);
-	  /* FIXME: should putback() here */
-	  /* FALLTHRU */
+	  /* FIXME: should putback() here? */
+	  break;
      case ';':
 	  break;
      case ',':
@@ -771,11 +884,24 @@ parse_variable_declaration(Ident *ident, int parm)
 	  goto again;
      case '=':
 	  nexttoken();
-	  if (tok.type == LBRACE || tok.type == LBRACE0)
+	  switch (tok.type) {
+	  case LBRACE:
+	  case LBRACE0:
 	       initializer_list();
-	  else
+	       goto select;
+
+	  case RBRACE:
+	  case RBRACE0:
+	       if (verbose)
+		    file_error(_("expected expression"), &tok);
+	       break;
+
+	  default:
 	       expression();
-	  goto select;
+	       goto select;
+	  }
+	  break;
+
      case LBRACE0:
      case LBRACE:
 	  func_body();
@@ -821,7 +947,7 @@ void
 parse_knr_dcl(Ident *ident)
 {
      ident->type_end = -1;
-     parse_dcl(ident, !strict_ansi);     
+     parse_dcl(ident, !strict_ansi);
 }
 
 void
@@ -831,16 +957,16 @@ skip_struct()
 	  nexttoken();
      } else if (tok.type == ';')
 	  return;
-     
+
      if (tok.type == LBRACE || tok.type == LBRACE0) {
-	  if (skip_balanced('{', '}', 1) == -1) {
+	  if (skip_balanced('{', 1) == -1) {
 	       file_error(_("unexpected end of file in struct"), NULL);
 	       return;
 	  }
      }
 
      while (tok.type == PARM_WRAPPER) {
-	  if (skip_balanced('(', ')', 0) == -1) {
+	  if (skip_balanced('(', 0) == -1) {
 	      file_error(_("unexpected end of file in struct"), NULL);
 	      return;
 	  }
@@ -851,19 +977,19 @@ void
 parse_typedef()
 {
      Ident ident;
-     
+
      ident.name = NULL;
      ident.type_end = -1;
      ident.parmcnt = -1;
      ident.line = -1;
      ident.storage = AnyStorage;
-     
+
      nexttoken();
      if (!fake_struct(&ident))
 	  putback();
-     
+
      dcl(&ident);
-     if (ident.name) 
+     if (ident.name)
 	  declare_type(&ident);
 }
 
@@ -877,7 +1003,7 @@ parse_dcl(Ident *ident, int maybe_knr)
      save_stack();
      if (ident->name)
 	  declare(ident, maybe_knr);
-     else 
+     else
 	  undo_save_stack();
 }
 
@@ -889,14 +1015,15 @@ dcl(Ident *idptr)
 	       if (idptr && idptr->type_end == -1)
 		    idptr->type_end = curs-1;
 	  } else if (tok.type == PARM_WRAPPER) {
-	       if (skip_balanced('(', ')', 0) == -1) {
+	       if (skip_balanced('(', 0) == -1) {
 		    file_error(_("unexpected end of file in function declaration"),
 			       NULL);
 		    return 1;
 	       }
+	       putback();
 	  } else if (tok.type == IDENTIFIER) {
 	       int type;
-	       
+
 	       while (tok.type == IDENTIFIER)
 		    nexttoken();
 	       type = tok.type;
@@ -913,17 +1040,72 @@ dcl(Ident *idptr)
      return dirdcl(idptr);
 }
 
+int
+getident(Ident *idptr, int **parm_ptr)
+{
+     Stackpos pos;
+     int bc = 0;
+
+     mark(pos);
+     do {
+	  ++bc;
+	  if (nexttoken() == 0) {
+	       if (verbose)
+		    file_error(_("unexpected end of file"), NULL);
+	       return -1;
+	  }
+     } while (tok.type == '(');
+
+     if (tok.type == IDENTIFIER) {
+	  char *name = tok.token;
+	  int line = tok.line;
+
+	  while (bc > 0) {
+	       if (nexttoken() == 0) {
+		    if (verbose)
+			 file_error(_("unexpected end of file"), NULL);
+		    return -1;
+	       }
+	       if (tok.type == ')')
+		    --bc;
+	       else
+		    break;
+	  }
+
+	  if (bc == 0) {
+	       if (idptr) {
+		    idptr->name = name;
+		    idptr->line = line;
+		    *parm_ptr = &idptr->parmcnt;
+	       }
+	       return 0;
+	  }
+     }
+     restore(pos);
+     return 1;
+}
+
 int
 dirdcl(Ident *idptr)
 {
      int wrapper = 0;
      int *parm_ptr = NULL;
-     
+
      if (tok.type == '(') {
-	  dcl(idptr);
-	  if (tok.type != ')' && verbose) {
-	       file_error(_("expected `)'"), &tok);
+	  switch (getident(idptr, &parm_ptr)) {
+	  case 0:
+	       break;
+
+	  case -1:
 	       return 1;
+
+	  case 1:
+	       dcl(idptr);
+	       if (tok.type != ')' && verbose) {
+		    file_error(_("expected `)'"), &tok);
+		    return 1;
+	       }
+	       break;
 	  }
      } else if (tok.type == IDENTIFIER) {
 	  if (idptr) {
@@ -932,7 +1114,7 @@ dirdcl(Ident *idptr)
 	       parm_ptr = &idptr->parmcnt;
 	  }
      }
-     
+
      if (nexttoken() == PARM_WRAPPER) {
 	  wrapper = 1;
 	  nexttoken(); /* read '(' */
@@ -940,9 +1122,13 @@ dirdcl(Ident *idptr)
 	  putback();
 
      while (nexttoken() == '[' || tok.type == '(') {
-	  if (tok.type == '[') 
-	       skip_to(']');
-	  else {
+	  if (tok.type == '[') {
+	       if (find_closing_paren('[', 1) == -1) {
+		    file_error(_("unexpected end of file in declaration"),
+			       NULL);
+		    return -1;
+	       }
+	  } else {
 	       maybe_parm_list(parm_ptr);
 	       if (tok.type != ')' && verbose) {
 		    file_error(_("expected `)'"), &tok);
@@ -954,11 +1140,11 @@ dirdcl(Ident *idptr)
 	  nexttoken(); /* read ')' */
 
      while (tok.type == PARM_WRAPPER) {
-	  if (skip_balanced('(', ')', 0) == -1)
+	  if (skip_balanced('(', 0) == -1)
 	       file_error(_("unexpected end of file in function declaration"),
 			  NULL);
      }
-     
+
      return 0;
 }
 
@@ -976,9 +1162,9 @@ parmdcl(Ident *idptr)
 		    nexttoken();
 	       type = tok.type;
 	       putback();
-	       if (type != MODIFIER) 
+	       if (type != MODIFIER)
 		    break;
-	  } else if (tok.type == ')' || tok.type == ',') 
+	  } else if (tok.type == ')' || tok.type == ',')
 	       return 0;
      }
      if (idptr && idptr->type_end == -1)
@@ -1021,7 +1207,7 @@ maybe_parm_list(int *parm_cnt_return)
 			       &tok);
 	       level = 0;
 	       do {
-		    if (tok.type == '(') 
+		    if (tok.type == '(')
 			 level++;
 		    else if (tok.type == ')') {
 			 if (level-- == 0)
@@ -1040,7 +1226,7 @@ void
 func_body()
 {
      Ident ident;
-     
+
      level++;
      move_parms(level);
      while (level) {
@@ -1065,7 +1251,7 @@ func_body()
 	       parse_declaration(&ident, 0);
 	       break;
 	  case LBRACE0:
-	  case '{':
+	  case LBRACE:
 	       level++;
 	       break;
 	  case RBRACE0:
@@ -1079,7 +1265,7 @@ func_body()
 	       }
 	       /* else: */
 	       /* FALLTHRU */
-	  case '}':
+	  case RBRACE:
 	       delete_autos(level);
 	       level--;
 	       break;
@@ -1087,10 +1273,10 @@ func_body()
 	       if (verbose)
 		    file_error(_("unexpected end of file in function body"),
 			       NULL);
-	       caller = NULL;
-	       return;
+	       goto end;
 	  }
      }
+end:
      caller = NULL;
 }
 
@@ -1102,14 +1288,15 @@ get_knr_args(Ident *ident)
      Ident id;
 
      switch (tok.type) {
+     case QUALIFIER:
      case IDENTIFIER:
      case TYPE:
      case STRUCT:
 	  /* maybe K&R function definition */
-	  
+
 	  mark(sp);
 	  parmcnt = 0;
-	  
+
 	  for (stop = 0; !stop && parmcnt < ident->parmcnt;
 	       nexttoken()) {
 	       id.type_end = -1;
@@ -1120,6 +1307,7 @@ get_knr_args(Ident *ident)
 		    stop = 1;
 		    break;
 	       case TYPE:
+	       case QUALIFIER:
 	       case IDENTIFIER:
 	       case STRUCT:
 		    putback();
@@ -1151,7 +1339,7 @@ void
 declare(Ident *ident, int maybe_knr)
 {
      Symbol *sp;
-     
+
      if (ident->storage == AutoStorage) {
 	  undo_save_stack();
 	  sp = install_ident(ident->name, ident->storage);
@@ -1162,7 +1350,7 @@ declare(Ident *ident, int maybe_knr)
 	       sp->level = level;
 	  sp->arity = -1;
 	  return;
-     } 
+     }
 
      if ((ident->parmcnt >= 0
 	  && (!maybe_knr || get_knr_args(ident) == 0)
@@ -1174,7 +1362,7 @@ declare(Ident *ident, int maybe_knr)
 	  /* add_external()?? */
 	  return;
      }
-     
+
      sp = get_symbol(ident->name);
      if (sp->source) {
 	  if (ident->storage == StaticStorage
@@ -1182,11 +1370,11 @@ declare(Ident *ident, int maybe_knr)
 	       sp = install_ident(ident->name, ident->storage);
 	  } else {
 	       if (sp->arity >= 0)
-		    error_at_line(0, 0, filename, ident->line, 
+		    error_at_line(0, 0, filename, ident->line,
 				  _("%s/%d redefined"),
 				  ident->name, sp->arity);
 	       else
-		    error_at_line(0, 0, filename, ident->line, 
+		    error_at_line(0, 0, filename, ident->line,
 				  _("%s redefined"),
 				  ident->name);
 	       error_at_line(0, 0, sp->source, sp->def_line,
@@ -1196,7 +1384,7 @@ declare(Ident *ident, int maybe_knr)
 
      sp->type = SymIdentifier;
      sp->arity = ident->parmcnt;
-     ident_change_storage(sp, 
+     ident_change_storage(sp,
 			  (ident->storage == ExplicitExternStorage) ?
 			  ExternStorage : ident->storage);
      sp->decl = finish_save_stack(ident->name);
@@ -1215,7 +1403,7 @@ void
 declare_type(Ident *ident)
 {
      Symbol *sp;
-     
+
      undo_save_stack();
      sp = lookup(ident->name);
      for ( ; sp; sp = sp->next)
@@ -1237,14 +1425,21 @@ Symbol *
 get_symbol(char *name)
 {
      Symbol *sp = lookup(name) ;
-     
+
      if (sp) {
 	  for (; sp; sp = sp->next) {
-	       if (sp->type == SymIdentifier && strcmp(sp->name, name) == 0)
+	       if (sp->flag == symbol_start ||
+		   sp->flag == symbol_target ||
+		   sp->type == SymIdentifier)
 		    break;
 	  }
-	  if (sp)
+	  if (sp) {
+	       if ((sp->flag == symbol_start || sp->flag == symbol_target)
+		   && sp->type == SymUndefined) {
+		    init_ident(sp, ExternStorage);
+	       }
 	       return sp;
+	  }
      }
      return install_ident(name, ExternStorage);
 }
@@ -1300,3 +1495,13 @@ reference(char *name, int line)
      }
 }
 
+/* If the caller is a static symbol, reset it to NULL.  This function
+   is called as a part of cleanup routine at the end of compilation
+   module. */
+void
+reset_static_caller(void)
+{
+     if (caller &&
+	 (caller->storage == StaticStorage || caller->flag == symbol_local))
+	  caller = NULL;
+}
