@@ -1,5 +1,5 @@
 /* This file is part of GNU cflow
-   Copyright (C) 1997-2021 Sergey Poznyakoff
+   Copyright (C) 1997-2025 Sergey Poznyakoff
 
    GNU cflow is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -28,12 +28,18 @@ static struct linked_list *static_func_list;
 static struct linked_list *unit_local_list;
 
 static void
-append_symbol(struct linked_list **plist, Symbol *sp)
+symbol_unlink_from_list(Symbol *sp)
 {
      if (sp->entry) {
 	  linked_list_unlink(sp->entry->list, sp->entry);
 	  sp->entry = NULL;
      }
+}
+
+static void
+append_symbol(struct linked_list **plist, Symbol *sp)
+{
+     symbol_unlink_from_list(sp);
      if (!data_in_list(sp, *plist)) {
 	  linked_list_append(plist, sp);
 	  sp->entry = (*plist)->tail;
@@ -146,6 +152,7 @@ ident_change_storage(Symbol *sp, enum storage storage)
 	  append_symbol(&auto_symbol_list, sp);
 	  break;
      default:
+	  symbol_unlink_from_list(sp);
 	  break;
      }
      
