@@ -1,5 +1,5 @@
 /* wordsplit - a word splitter
-   Copyright (C) 2009-2019 Sergey Poznyakoff
+   Copyright (C) 2009-2025 Sergey Poznyakoff
 
    This program is free software; you can redistribute it and/or modify it
    under the terms of the GNU General Public License as published by the
@@ -67,6 +67,8 @@ is_name_char (struct wordsplit *wsp, int c)
 #define to_num(c) \
   (ISDIGIT(c) ? c - '0' : (ISXDIGIT(c) ? toupper(c) - 'A' + 10 : 255 ))
 
+static int wsplt_unquote_char (const char *transtab, int c);
+
 #define ALLOC_INIT 128
 #define ALLOC_INCR 128
 
@@ -183,7 +185,7 @@ _wsplt_subsplit (struct wordsplit *wsp, struct wordsplit *wss,
 
   wss->ws_options = wsp->ws_options & ~WRDSO_MAXWORDS;
   wss->ws_namechar = wsp->ws_namechar;
-    
+
   flags |= WRDSF_DELIM
 	 | WRDSF_ALLOC_DIE
 	 | WRDSF_ERROR
@@ -247,7 +249,16 @@ wordsplit_init0 (struct wordsplit *wsp)
   wsp->ws_errno = 0;
 }
 
-char wordsplit_c_escape_tab[] = "\\\\\"\"a\ab\bf\fn\nr\rt\tv\v";
+char const *wordsplit_escape[] = {
+	/* C-style escapes, for quoted strings */
+	[WS_ESC_C]     = "\\\\\"\"a\ab\bf\fn\nr\rt\tv\v",
+	/* C-style escapes, outsize of quoted strings */
+	[WS_ESC_C_WS]  = "\\\\\"\"a\ab\bf\fn\nr\rt\tv\v  \t\t",
+	/* Escape double-quote and backslash. */
+	[WS_ESC_DQ]    = "\\\\\"\"",
+	/* Escape double-quote, backslash, and whitespace. */
+	[WS_ESC_DQ_WS] = "\\\\\"\"  \t\t"
+};
 
 static int
 wordsplit_init (struct wordsplit *wsp, const char *input, size_t len,
@@ -314,21 +325,17 @@ wordsplit_init (struct wordsplit *wsp, const char *input, size_t len,
       if (!wsp->ws_escape[WRDSX_QUOTE])
 	wsp->ws_escape[WRDSX_QUOTE] = "";
     }
+  else if (wsp->ws_flags & WRDSF_CESCAPES)
+    {
+      wsp->ws_escape[WRDSX_WORD] = wordsplit_escape[WS_ESC_C_WS];
+      wsp->ws_escape[WRDSX_QUOTE] = wordsplit_escape[WS_ESC_C];
+      wsp->ws_options |= WRDSO_OESC_QUOTE | WRDSO_OESC_WORD
+		      | WRDSO_XESC_QUOTE | WRDSO_XESC_WORD;
+    }
   else
     {
-      if (wsp->ws_flags & WRDSF_CESCAPES)
-	{
-	  wsp->ws_escape[WRDSX_WORD] = wordsplit_c_escape_tab;
-	  wsp->ws_escape[WRDSX_QUOTE] = wordsplit_c_escape_tab;
-	  wsp->ws_options |= WRDSO_OESC_QUOTE | WRDSO_OESC_WORD
-			     | WRDSO_XESC_QUOTE | WRDSO_XESC_WORD;
-	}
-      else
-	{
-	  wsp->ws_escape[WRDSX_WORD] = "";
-	  wsp->ws_escape[WRDSX_QUOTE] = "\\\\\"\"";
-	  wsp->ws_options |= WRDSO_BSKEEP_QUOTE;
-	}
+      wsp->ws_escape[WRDSX_WORD] = "";
+      wsp->ws_escape[WRDSX_QUOTE] = "";
     }
 
   if (!(wsp->ws_options & WRDSO_PARAMV))
@@ -346,7 +353,7 @@ wordsplit_init (struct wordsplit *wsp, const char *input, size_t len,
     }
   else
     wsp->ws_namechar = NULL;
-  
+
   wsp->ws_endp = 0;
   wsp->ws_wordi = 0;
 
@@ -700,14 +707,8 @@ wsnode_quoteremoval (struct wordsplit *wsp)
     {
       const char *str = wsnode_ptr (wsp, p);
       size_t slen = wsnode_len (p);
-      int unquote;
 
-      if (wsp->ws_flags & WRDSF_QUOTE)
-	unquote = !(p->flags & _WSNF_NOEXPAND);
-      else
-	unquote = 0;
-
-      if (unquote)
+      if (!(p->flags & _WSNF_NOEXPAND))
 	{
 	  if (!(p->flags & _WSNF_WORD))
 	    {
@@ -765,7 +766,6 @@ wordsplit_finish (struct wordsplit *wsp)
 {
   struct wordsplit_node *p;
   size_t n;
-  int delim;
 
   /* Postprocess delimiters. It would be rather simple, if it weren't for
      the incremental operation.
@@ -793,7 +793,6 @@ wordsplit_finish (struct wordsplit *wsp)
 	is set.
   */
  again:
-  delim = 0;         /* Delimiter being processed (if any) */
   n = 0;             /* Number of words processed so far */
   p = wsp->ws_head;  /* Current node */
 
@@ -802,56 +801,48 @@ wordsplit_finish (struct wordsplit *wsp)
       struct wordsplit_node *next = p->next;
       if (p->flags & _WSNF_DELIM)
 	{
-	  if (wsp->ws_flags & WRDSF_RETURN_DELIMS)
+	  if (wsp->ws_flags & WRDSF_SQUEEZE_DELIMS)
 	    {
-	      if (wsp->ws_flags & WRDSF_SQUEEZE_DELIMS)
+	      if (next)
 		{
-		  char const *s = wsnode_ptr (wsp, p);
-		  if (delim)
+		  if ((next->flags & _WSNF_DELIM) &&
+		      (wsnode_ptr (wsp, p))[0] == (wsnode_ptr (wsp, next))[0])
 		    {
-		      if (delim == *s)
-			{
-			  wsnode_remove (wsp, p);
-			  p = next;
-			  continue;
-			}
-		      else
-			{
-			  delim = 0;
-			  n++; /* Count this node; it will be returned */
-			}
-		    }
-		  else
-		    {
-		      delim = *s;
+		      wsnode_remove (wsp, p);
 		      p = next;
 		      continue;
 		    }
 		}
+	      else if (wsp->ws_flags & WRDSF_INCREMENTAL)
+		goto restart;
 	    }
-	  else if (wsp->ws_options & WRDSO_MAXWORDS)
+	  else if ((next && (next->flags & _WSNF_DELIM)) &&
+		   (!(wsp->ws_options & WRDSO_RETDELNOTEMPTY)))
+	    {
+	      int rc;
+	      struct wordsplit_node *nulnode;
+	      if ((rc = wsnode_new (wsp, &nulnode)) != 0)
+		return rc;
+	      nulnode->flags = _WSNF_NULL | _WSNF_NOEXPAND;
+	      wsnode_insert (wsp, nulnode, p, 0);
+	      next = nulnode;
+	    }
+
+	  if ((wsp->ws_options & WRDSO_MAXWORDS) &&
+	      !(wsp->ws_flags & WRDSF_RETURN_DELIMS))
 	    {
 	      wsnode_remove (wsp, p);
 	      p = next;
 	      continue;
 	    }
 	}
-      else
-	{
-	  if (delim)
-	    {
-	      /* Last node was a delimiter or a compressed run of delimiters;
-		 Count it, and clear the delimiter marker */
-	      n++;
-	      delim = 0;
-	    }
-	  if (wsp->ws_options & WRDSO_MAXWORDS)
-	    {
-	      if (wsp->ws_wordi + n + 1 == wsp->ws_maxwords)
-		break;
-	    }
-	}
+
       n++;
+
+      if ((wsp->ws_options & WRDSO_MAXWORDS) &&
+	  (wsp->ws_wordi + n == wsp->ws_maxwords))
+	  break;
+
       if (wsp->ws_flags & WRDSF_INCREMENTAL)
 	p = NULL; /* Break the loop */
       else
@@ -874,6 +865,7 @@ wordsplit_finish (struct wordsplit *wsp)
       if (wsp->ws_flags & WRDSF_INCREMENTAL)
 	{
 	  /* Restart the processing, if there's any input left. */
+	restart:
 	  if (wsp->ws_endp < wsp->ws_len)
 	    {
 	      int rc;
@@ -1071,7 +1063,7 @@ wsplt_env_find (struct wordsplit *wsp, const char *name, size_t len)
 	    break;
 	}
     }
-  else 
+  else
     {
       /* Usual (A=B) environment. */
       for (i = 0; wsp->ws_env[i]; i++)
@@ -1442,7 +1434,7 @@ expvar (struct wordsplit *wsp, const char *str, size_t len,
 				   && (str[1] == '-'
 				       && ISDIGIT (str[2]))))) != 0))
     {
-      int i0 = str[0] == '-' ? 1 : 0;
+      int i0 = str[1] == '-' ? 1 : 0;
       str++;
       len--;
       for (i = i0; i < len; i++)
@@ -1679,7 +1671,7 @@ expvar (struct wordsplit *wsp, const char *str, size_t len,
 	    return 1;
 	  wsnode_insert (wsp, newnode, *ptail, 0);
 	  *ptail = newnode;
-	  newnode->flags = _WSNF_NULL;
+	  newnode->flags = _WSNF_NULL | _WSNF_NOEXPAND;
 	}
       else
 	{
@@ -1725,7 +1717,7 @@ expvar (struct wordsplit *wsp, const char *str, size_t len,
 	return 1;
       wsnode_insert (wsp, newnode, *ptail, 0);
       *ptail = newnode;
-      newnode->flags = _WSNF_NULL;
+      newnode->flags = _WSNF_NULL | _WSNF_NOEXPAND;
     }
   return 0;
 }
@@ -1896,7 +1888,7 @@ expcmd (struct wordsplit *wsp, const char *str, size_t len,
 	    return 1;
 	  wsnode_insert (wsp, newnode, *ptail, 0);
 	  *ptail = newnode;
-	  newnode->flags = _WSNF_NULL;
+	  newnode->flags = _WSNF_NULL | _WSNF_NOEXPAND;
 	}
       else
 	{
@@ -1927,7 +1919,7 @@ expcmd (struct wordsplit *wsp, const char *str, size_t len,
 	return 1;
       wsnode_insert (wsp, newnode, *ptail, 0);
       *ptail = newnode;
-      newnode->flags = _WSNF_NULL;
+      newnode->flags = _WSNF_NULL | _WSNF_NOEXPAND;
     }
   return 0;
 }
@@ -1982,7 +1974,7 @@ wordsplit_trimws (struct wordsplit *wsp)
 	   n > p->v.segm.beg && ISWS (wsp->ws_input[n - 1]); n--);
       p->v.segm.end = n;
       if (p->v.segm.beg == p->v.segm.end)
-	p->flags |= _WSNF_NULL;
+	p->flags |= _WSNF_NULL | _WSNF_NOEXPAND;
     }
 
   wsnode_nullelim (wsp);
@@ -2148,7 +2140,7 @@ wordsplit_pathexpand (struct wordsplit *wsp)
 
 	    default:
 	      free (pattern);
-	      return _wsplt_setctxerr (wsp, WRDSE_GLOBERR, pattern, slen);
+	      return _wsplt_setctxerr (wsp, WRDSE_GLOBERR, str, slen);
 	    }
 
 	  prev = p;
@@ -2303,28 +2295,32 @@ scan_word (struct wordsplit *wsp, size_t start, int consume_all)
 	      return _WRDS_OK;
 	    }
 
-	  if (wsp->ws_flags & WRDSF_QUOTE)
+	  if (command[i] == '\\')
 	    {
-	      if (command[i] == '\\')
+	      if (i + 1 == len)
 		{
-		  if (++i == len)
-		    break;
 		  i++;
+		  break;
+		}
+	      if (wsplt_unquote_char (wsp->ws_escape[WRDSX_WORD], command[i+1]))
+		{
+		  i += 2;
 		  continue;
 		}
+	    }
 
-	      if (((wsp->ws_flags & WRDSF_SQUOTE) && command[i] == '\'') ||
-		  ((wsp->ws_flags & WRDSF_DQUOTE) && command[i] == '"'))
-		{
-		  if (join && wsp->ws_tail)
-		    wsp->ws_tail->flags |= _WSNF_JOIN;
-		  if (wordsplit_add_segm (wsp, start, i, _WSNF_JOIN))
-		    return _WRDS_ERR;
-		  if (scan_qstring (wsp, i, &i))
-		    return _WRDS_ERR;
-		  start = i + 1;
-		  join = 1;
-		}
+	  if ((wsp->ws_flags & WRDSF_QUOTE) &&
+	      (((wsp->ws_flags & WRDSF_SQUOTE) && command[i] == '\'') ||
+	       ((wsp->ws_flags & WRDSF_DQUOTE) && command[i] == '"')))
+	    {
+	      if (join && wsp->ws_tail)
+		wsp->ws_tail->flags |= _WSNF_JOIN;
+	      if (wordsplit_add_segm (wsp, start, i, _WSNF_JOIN))
+		return _WRDS_ERR;
+	      if (scan_qstring (wsp, i, &i))
+		return _WRDS_ERR;
+	      start = i + 1;
+	      join = 1;
 	    }
 
 	  if (command[i] == '$')
@@ -2350,7 +2346,7 @@ scan_word (struct wordsplit *wsp, size_t start, int consume_all)
   else if (WSP_RETURN_DELIMS (wsp))
     {
       i++;
-      flags |= _WSNF_DELIM;
+      flags |= _WSNF_DELIM | _WSNF_NOEXPAND;
     }
   else if (!(wsp->ws_flags & WRDSF_SQUEEZE_DELIMS))
     flags |= _WSNF_EMPTYOK;
@@ -2449,13 +2445,13 @@ wsplt_quote_char (const char *transtab, int c)
 int
 wordsplit_c_unquote_char (int c)
 {
-  return wsplt_unquote_char (wordsplit_c_escape_tab, c);
+  return wsplt_unquote_char (wordsplit_escape[WS_ESC_C], c);
 }
 
 int
 wordsplit_c_quote_char (int c)
 {
-  return wsplt_quote_char (wordsplit_c_escape_tab, c);
+  return wsplt_quote_char (wordsplit_escape[WS_ESC_C], c);
 }
 
 void
@@ -2708,7 +2704,11 @@ wordsplit_run (const char *command, size_t length, struct wordsplit *wsp,
   if (!command)
     {
       if (!(flags & WRDSF_INCREMENTAL))
-	return _wsplt_seterr (wsp, WRDSE_USAGE);
+	{
+	  /* Make sure _wsplt_seterr gets correct flags. */
+	  wsp->ws_flags = flags;
+	  return _wsplt_seterr (wsp, WRDSE_USAGE);
+	}
 
       if (wsp->ws_head)
 	return wordsplit_finish (wsp);
